@@ -1,6 +1,6 @@
 ##############################################################################
 #
-# Copyright (C) Zenoss, Inc. 2013, all rights reserved.
+# Copyright (C) Zenoss, Inc. 2016-2017, all rights reserved.
 #
 # This content is made available according to terms specified in the LICENSE
 # file at the top-level directory of this package.
@@ -11,15 +11,11 @@ import copy
 import logging
 from collections import namedtuple
 from httplib import BAD_REQUEST, UNAUTHORIZED, FORBIDDEN, OK
-import shlex
-from cStringIO import StringIO
 
-from twisted.internet import reactor
 from twisted.internet.defer import (
     inlineCallbacks,
     returnValue,
-    DeferredSemaphore,
-    succeed
+    DeferredSemaphore
 )
 from twisted.internet.error import TimeoutError
 
@@ -50,7 +46,6 @@ from .util import (
     ET,
 )
 from .shell import (
-    _find_shell_id,
     _build_command_line_elem,
     _build_ps_command_line_elem,
     _find_command_id,
@@ -59,17 +54,35 @@ from .shell import (
     _find_exit_code,
     CommandResponse,
     _stripped_lines,
+    _find_shell_id
 )
 from .enumerate import (
     DEFAULT_RESOURCE_URI,
     SaxResponseHandler,
-    _MAX_REQUESTS_PER_ENUMERATION
+    _MAX_REQUESTS_PER_ENUMERATION,
+    ItemsAccumulator
 )
 from .SessionManager import SESSION_MANAGER, Session
+from .twisted_utils import with_timeout
 kerberos = None
 LOG = logging.getLogger('winrm')
 
 
+def create_shell_from_elem(elem):
+    accumulator = ItemsAccumulator()
+    accumulator.new_item()
+    for item in ['ShellId', 'Owner', 'ClientIP', 'ShellRunTime', 'ShellInactivity', 'IdleTimeOut']:
+        xpath = './/{{{}}}{}'.format(c.XML_NS_MSRSP, item)
+        try:
+            accumulator.add_property(item, elem.findtext(xpath).strip())
+        except AttributeError as e:
+            if item == 'ShellId':
+                raise Exception('Invalid response from create shell request: {}'.format(e))
+            # as long as we have a valid ShellId we should be fine
+            accumulator.add_property(item, '')
+    return accumulator.items[0]
+
+
 class WinRMSession(Session):
     '''
     Session class to keep track of single winrm connection
@@ -97,8 +110,18 @@ class WinRMSession(Session):
         # connection.
         self.sem = DeferredSemaphore(1)
 
+        # unused.  reserved for possible future use
         self._refresh_dc = None
 
+    def semrun(self, fn, *args, **kwargs):
+        """Run fn(*args, **kwargs) under a DeferredSemaphore with a timeout."""
+        return self.sem.run(
+            with_timeout,
+            fn=fn,
+            args=args,
+            kwargs=kwargs,
+            seconds=self._conn_info.timeout)
+
     def is_kerberos(self):
         return self._conn_info.auth_type == 'kerberos'
 
@@ -118,10 +141,13 @@ class WinRMSession(Session):
             self._headers.addRawHeader('Connection', self._conn_info.connectiontype)
         return self._headers
 
+    def update_conn_info(self, client):
+        self._conn_info = client._conn_info
+
     @inlineCallbacks
     def _deferred_login(self, client=None):
         if client:
-            self._conn_info = client._conn_info
+            self.update_conn_info(client)
         self._url = "{c.scheme}://{c.ipaddress}:{c.port}/wsman".format(c=self._conn_info)
         if self.is_kerberos():
             self._gssclient = yield _authenticate_with_kerberos(self._conn_info, self._url, self._agent)
@@ -130,30 +156,42 @@ class WinRMSession(Session):
             returnValue('basic_auth_token')
 
     @inlineCallbacks
-    def _deferred_logout(self):
+    def close_cached_connections(self):
         # close connections so we do not end up with orphans
         # return a Deferred()
-        self.loggedout = True
         if self._agent and hasattr(self._agent, 'closeCachedConnections'):
             # twisted 11 has no return and is part of the Agent
-            return succeed(self._agent.closeCachedConnections())
+            self._agent.closeCachedConnections()
         elif self._agent:
-            # twisted 12 returns a Deferred
-            return self._agent._pool.closeCachedConnections()
-        else:
-            # no agent
-            return succeed(None)
+            # twisted 12 has a pool
+            yield self._agent._pool.closeCachedConnections()
+        returnValue(None)
+
+    @inlineCallbacks
+    def _deferred_logout(self):
+        # close connections so they don't timeout
+        # gssclient will no longer be valid so get rid of it
+        # set token to None so the next client will reinitialize
+        #   the connection
+        if self._gssclient is not None:
+            self._gssclient.cleanup()
+            self._gssclient = None
+        self._token = None
+        yield self.close_cached_connections()
+        returnValue(None)
 
     @inlineCallbacks
     def handle_response(self, request, response, client):
         if response.code == UNAUTHORIZED or response.code == BAD_REQUEST:
             # check to see if we need to re-authorize due to lost connection or bad request error
+            # only retry if using kerberos
+            self._token = None
+            yield self.close_cached_connections()
+            self._login_d = None
             if self._gssclient is not None:
                 self._gssclient.cleanup()
                 self._gssclient = None
                 self._token = None
-                self._agent = _get_agent()
-                self._login_d = None
                 yield SESSION_MANAGER.init_connection(client, WinRMSession)
                 try:
                     yield self._set_headers()
@@ -190,8 +228,11 @@ class WinRMSession(Session):
                 reader = _ErrorReader()
             response.deliverBody(reader)
             message = yield reader.d
-            raise RequestError("HTTP status: {}. {}".format(
-                response.code, message))
+            if 'maximum number of concurrent operations for this user has been exceeded' in message:
+                message += '  To fix this, increase the MaxConcurrentOperationsPerUser WinRM'\
+                           ' Configuration option to 4294967295 and restart the winrm service.'
+            raise RequestError("{}: HTTP status: {}. {}.".format(
+                client._conn_info.ipaddress, response.code, message))
         returnValue(response)
 
     @inlineCallbacks
@@ -217,6 +258,12 @@ class WinRMSession(Session):
     @inlineCallbacks
     def _send_request(self, request_template_name, client, envelope_size=None,
                       locale=None, code_page=None, **kwargs):
+        if self._logout_dc is not None:
+            try:
+                self._logout_dc.cancel()
+                self._logout_dc = None
+            except Exception:
+                pass
         kwargs['envelope_size'] = envelope_size or client._conn_info.envelope_size
         kwargs['locale'] = locale or client._conn_info.locale
         kwargs['code_page'] = code_page or client._conn_info.code_page
@@ -245,16 +292,6 @@ class WinRMSession(Session):
         response = yield self.handle_response(request, response, client)
         returnValue(response)
 
-    def close_connection(self, client):
-        try:
-            self._refresh_dc.cancel()
-        except Exception:
-            pass
-
-        # Close the connection after 60 seconds.  This will give other clients
-        # enough time to keep the connection alive and continue using the same session.
-        self._refresh_dc = reactor.callLater(60, SESSION_MANAGER.close_connection, client)
-
 
 class WinRMClient(object):
     """Base winrm client class
@@ -266,18 +303,17 @@ class WinRMClient(object):
         self.key = None
         self._conn_info = conn_info
         self.ps_script = None
+        self._shell_id = None
+        self._session = None
 
-    def _session(self):
-        return SESSION_MANAGER.get_connection(self.key)
+    def session(self):
+        if self._session is None:
+            self._session = SESSION_MANAGER.get_connection(self.key)
+        return self._session
 
     @inlineCallbacks
     def init_connection(self):
         """Initialize a connection through the session_manager"""
-        if self._session() is not None:
-            try:
-                self._refresh_dc.cancel()
-            except Exception:
-                pass
         yield SESSION_MANAGER.init_connection(self, WinRMSession)
         returnValue(None)
 
@@ -285,22 +321,24 @@ class WinRMClient(object):
         return self._conn_info.auth_type == 'kerberos'
 
     def decrypt_body(self, body):
-        return self._session().decrypt_body(body)
+        return self.session().decrypt_body(body)
 
     @inlineCallbacks
     def send_request(self, request, **kwargs):
-        if not self._session():
+        if self.session() is None or self._session._token is None\
+                or (self._session.is_kerberos() and self._session._gssclient is None):
             yield self.init_connection()
 
-        if not self._session():
+        if not self.session():
             raise Exception('Could not connect to device {}'.format(self.conn_info.hostname))
-        response = yield self._session().send_request(request, self, **kwargs)
+        response = yield self.session().send_request(request, self, **kwargs)
         returnValue(response)
 
     @inlineCallbacks
     def _create_shell(self):
         elem = yield self.send_request('create')
-        returnValue(_find_shell_id(elem))
+        self._shell_id = _find_shell_id(elem)
+        returnValue(self._shell_id)
 
     @inlineCallbacks
     def _delete_shell(self, shell_id):
@@ -344,10 +382,12 @@ class WinRMClient(object):
             'receive', shell_id=shell_id, command_id=command_id)
         returnValue(receive_elem)
 
-    @inlineCallbacks
     def close_connection(self):
-        yield self._session().close_connection(self)
-        returnValue(None)
+        SESSION_MANAGER.close_connection(self)
+
+    @inlineCallbacks
+    def close_cached_connections(self):
+        yield self.session().close_cached_connections()
 
 
 class SingleCommandClient(WinRMClient):
@@ -370,10 +410,11 @@ class SingleCommandClient(WinRMClient):
         self.ps_script = ps_script
         yield self.init_connection()
         try:
-            cmd_response = yield self._session().sem.run(self.run_single_command,
-                                                         command_line)
+            cmd_response = yield self.session().semrun(
+                self.run_single_command,
+                command_line)
         except Exception:
-            yield self.close_connection()
+            self.close_connection()
             raise
         returnValue(cmd_response)
 
@@ -388,14 +429,13 @@ class SingleCommandClient(WinRMClient):
                 .stderr = [<non-empty, stripped line>, ...]
                 .exit_code = <int>
         """
-        shell_id = yield self._create_shell()
+        yield self._create_shell()
         cmd_response = None
         try:
-            cmd_response = yield self._run_command(shell_id, command_line)
+            cmd_response = yield self._run_command(self._shell_id, command_line)
         except TimeoutError:
-            yield self.close_connection()
-        yield self._delete_shell(shell_id)
-        yield self.close_connection()
+            yield self.close_cached_connections()
+        self.close_connection()
         returnValue(cmd_response)
 
     @inlineCallbacks
@@ -418,6 +458,7 @@ class SingleCommandClient(WinRMClient):
         yield self._signal_terminate(shell_id, command_id)
         stdout = _stripped_lines(stdout_parts)
         stderr = _stripped_lines(stderr_parts)
+        yield self._delete_shell(self._shell_id)
         returnValue(CommandResponse(stdout, stderr, exit_code))
 
 
@@ -450,14 +491,13 @@ class LongCommandClient(WinRMClient):
         self.key = (self._conn_info.ipaddress, command_line + str(ps_script))
         self.ps_script = ps_script
         yield self.init_connection()
-        if self._shell_id is None:
-            self._shell_id = yield self._create_shell()
+        self._shell_id = yield self._create_shell()
         try:
             command_elem = yield self._send_command(self._shell_id,
                                                     command_line)
         except TimeoutError:
             yield self._sender.send_request('delete', shell_id=self._shell_id)
-            yield self.close_connection()
+            self.close_connection()
             raise
         self._command_id = _find_command_id(command_elem)
         returnValue(None)
@@ -467,7 +507,7 @@ class LongCommandClient(WinRMClient):
         try:
             receive_elem = yield self._send_receive(self._shell_id, self._command_id)
         except TimeoutError:
-            yield self.close_connection()
+            self.close_connection()
             raise
         stdout_parts = _find_stream(receive_elem, self._command_id, 'stdout')
         stderr_parts = _find_stream(receive_elem, self._command_id, 'stderr')
@@ -486,7 +526,7 @@ class LongCommandClient(WinRMClient):
         yield self._signal_terminate(self._shell_id, self._command_id)
         yield self._delete_shell(self._shell_id)
         if close:
-            yield self.close_connection()
+            self.close_connection()
         returnValue(CommandResponse(stdout, stderr, self._exit_code))
 
 
@@ -501,12 +541,12 @@ class EnumerateClient(WinRMClient):
         super(EnumerateClient, self).__init__(conn_info)
         self._handler = SaxResponseHandler(self)
         self._hostname = conn_info.ipaddress
-        self.key = (conn_info.ipaddress, 'short')
+        self.key = (conn_info.ipaddress, 'enumerate')
 
     @inlineCallbacks
     def enumerate(self, wql, resource_uri=DEFAULT_RESOURCE_URI):
         """Runs a remote WQL query."""
-        if self._session() is None:
+        if self.session() is None:
             yield self.init_connection()
         request_template_name = 'enumerate'
         enumeration_context = None
@@ -515,7 +555,7 @@ class EnumerateClient(WinRMClient):
             for i in xrange(_MAX_REQUESTS_PER_ENUMERATION):
                 LOG.info('{0} "{1}" {2}'.format(
                     self._hostname, wql, request_template_name))
-                response = yield self._session()._send_request(
+                response = yield self.session()._send_request(
                     request_template_name,
                     self,
                     resource_uri=resource_uri,
@@ -549,9 +589,10 @@ class EnumerateClient(WinRMClient):
         yield self.init_connection()
         for enum_info in enum_infos:
             try:
-                items[enum_info] = yield self._session().sem.run(self.enumerate,
-                                                                 enum_info.wql,
-                                                                 enum_info.resource_uri)
+                items[enum_info] = yield self.session().semrun(
+                    self.enumerate,
+                    enum_info.wql,
+                    enum_info.resource_uri)
             except (UnauthorizedError, ForbiddenError):
                 # Fail the collection for general errors.
                 raise
@@ -559,7 +600,7 @@ class EnumerateClient(WinRMClient):
                 # Store empty results for other query-specific errors.
                 continue
 
-        yield self.close_connection()
+        self.close_connection()
         returnValue(items)
 
 
