@@ -18,8 +18,10 @@ from cStringIO import StringIO
 from twisted.internet import reactor, defer, task
 from twisted.internet.error import TimeoutError
 from xml.etree import cElementTree as ET
+from xml.etree import ElementTree
 from . import constants as c
 from .util import create_etree_request_sender, get_datetime, RequestError
+from .enumerate import create_parser_and_factory
 
 log = logging.getLogger('winrm')
 _MAX_REQUESTS_PER_COMMAND = 9999
@@ -214,36 +216,133 @@ def create_single_shot_command(conn_info):
     return SingleShotCommand(sender)
 
 
+def _find_enum_context(elem):
+    e_context = None
+    xpath = './/{{{}}}EnumerationContext'.format(c.XML_NS_ENUMERATION)
+    ctxt_elem = elem.find(xpath)
+    if ctxt_elem is not None:
+        e_context = ctxt_elem.text.split(':')[1]
+    return e_context
+
+
+def _find_shell_ids(elem):
+    ids = []
+    xpath = './/{{{}}}ShellId'.format(c.XML_NS_MSRSP)
+    shells = elem.findall(xpath)
+    for shell in shells:
+        ids.append(shell.text)
+    return ids
+
+
+@defer.inlineCallbacks
+def _get_active_shells(request_sender):
+    elem = yield request_sender.send_request('enum_shells')
+    enum_context = _find_enum_context(elem)
+    if enum_context is None:
+        defer.returnValue(None)
+    response = yield request_sender.send_request('pull_shells', uuid=enum_context)
+    body = ElementTree.tostring(response)
+    parser, factory = create_parser_and_factory()
+    parser.feed(body)
+    defer.returnValue(factory.items)
+
+
+@defer.inlineCallbacks
+def _get_active_shell(request_sender, conn_info, min_runtime=600):
+    """Sift through existing shells to find what should be the oldest active shell
+    created by our user.  Compare against minimum runtime so we grab the oldest
+    shell.  something less than min_runtime could have been created by a different
+    client.  sender can be RequestSender or WinRMClient.  conn_info is a
+    txwinrm.util.ConnectionInfo instance
+    """
+    shells = yield _get_active_shells(request_sender)
+    active_shell = None
+
+    user_domain = conn_info.username.split('@')
+    try:
+        # get domain user as netbios
+        user = (user_domain[1].split('.')[0] + '\\' + user_domain[0]).lower()
+    except IndexError:
+        # local user, no netbios
+        # user_domain[0] will always exist
+        user = user_domain[0].lower()
+
+    def get_runtime(runtime):
+        # return total runtime in seconds
+        # ShellRunTime is specified as P<days>DT<hours>H<minutes>M<seconds>S
+        # e.g. P1DT1H1M1S is a runtime of of 1 day, 1 hour, 1 minute, and 1 second
+        # we'll calculate the total number of seconds a shell has been running using
+        # these numbers
+        try:
+            rt_match = re.match('P(?P<d>\d+)DT(?P<h>\d+)H(?P<m>\d+)M(?P<s>\d+)S', runtime)
+            return int(rt_match.group('s')) + (int(rt_match.group('m')) * 60) + (int(rt_match.group('h')) * 3600) + (int(rt_match.group('d')) * 86400)
+        except Exception:
+            return 0
+
+    for shell in shells:
+        if user == shell.Owner.lower():
+            runtime = get_runtime(shell.ShellRunTime)
+            if runtime > min_runtime:
+                # found possible candidate, test against previous
+                if active_shell is not None:
+                    prev_runtime = get_runtime(active_shell.ShellRunTime)
+                    if runtime > prev_runtime:
+                        active_shell = shell
+                else:
+                    active_shell = shell
+    defer.returnValue(active_shell)
+
+
 class LongRunningCommand(object):
 
-    def __init__(self, sender):
+    def __init__(self, sender, min_runtime=600):
         self._sender = sender
         self._shell_id = None
         self._command_id = None
         self._exit_code = None
 
+        # attach to shell with a minimum runtime of x seconds to know
+        # that our user created the shell
+        self._min_runtime = min_runtime
+
+    @defer.inlineCallbacks
+    def is_shell_active(self, shell_id):
+        if shell_id is None:
+            defer.returnValue(False)
+        elem = yield self._sender.send_request('enum_shells')
+        enum_context = _find_enum_context(elem)
+        if enum_context is None:
+            defer.returnValue(False)
+        elem = yield self._sender.send_request('pull_shells', uuid=enum_context)
+        shell_ids = _find_shell_ids(elem)
+        if shell_id in shell_ids:
+            defer.returnValue(True)
+        else:
+            defer.returnValue(False)
+
     @defer.inlineCallbacks
     def start(self, command_line, ps_script=None):
-        log.debug("LongRunningCommand run_command: {0}".format(command_line))
-        if self._shell_id is None:
-            elem = yield self._sender.send_request('create')
-            self._shell_id = _find_shell_id(elem)
+        elem = yield self._sender.send_request('create')
+        self._shell_id = _find_shell_id(elem)
         if ps_script is not None:
+            log.debug("LongRunningCommand run_command: {0}".format(command_line + ps_script))
             command_line_elem = _build_ps_command_line_elem(command_line, ps_script)
         else:
+            log.debug("LongRunningCommand run_command: {0}".format(command_line))
             command_line_elem = _build_command_line_elem(command_line)
         log.debug('LongRunningCommand run_command: sending command request '
                   '(shell_id={0}, command_line_elem={1})'.format(
-                    self._shell_id, command_line_elem))
+                      self._shell_id, command_line_elem))
         try:
             command_elem = yield self._sender.send_request(
                 'command', shell_id=self._shell_id,
                 command_line_elem=command_line_elem,
                 timeout=self._sender._sender._conn_info.timeout)
         except TimeoutError:
-            yield self.delete_and_close()
+            yield self._sender.close_connections()
             raise
         self._command_id = _find_command_id(command_elem)
+        defer.returnValue(self._command_id)
 
     @defer.inlineCallbacks
     def receive(self):
@@ -253,8 +352,17 @@ class LongRunningCommand(object):
                 shell_id=self._shell_id,
                 command_id=self._command_id)
         except TimeoutError:
-            yield self.delete_and_close()
-            raise
+            # could be simple network problem, reconnect and try again
+            yield self._sender.close_connections()
+            try:
+                receive_elem = yield self._sender.send_request(
+                    'receive',
+                    shell_id=self._shell_id,
+                    command_id=self._command_id)
+            except TimeoutError:
+                yield self._sender.close_connections()
+            except Exception:
+                raise
         stdout_parts = _find_stream(receive_elem, self._command_id, 'stdout')
         stderr_parts = _find_stream(receive_elem, self._command_id, 'stderr')
         self._exit_code = _find_exit_code(receive_elem, self._command_id)
@@ -262,14 +370,6 @@ class LongRunningCommand(object):
         stderr = _stripped_lines(stderr_parts)
         defer.returnValue((stdout, stderr))
 
-    @defer.inlineCallbacks
-    def delete_and_close(self):
-        """Delete the remote shell and close connection"""
-        yield self._sender.send_request('delete', shell_id=self._shell_id)
-        self._shell_id = None
-        yield self._sender.close_connections()
-        defer.returnValue(None)
-
     @defer.inlineCallbacks
     def stop(self):
         for _ in xrange(_MAX_RETRIES):
@@ -279,23 +379,22 @@ class LongRunningCommand(object):
                     shell_id=self._shell_id,
                     command_id=self._command_id,
                     signal_code=c.SHELL_SIGNAL_CTRL_C)
-            except RequestError as e:
-                # if we get a 500 error back, let's try again
-                if 'HTTP status: 500. An internal error occurred' in e.message:
-                    continue
-                else:
-                    yield self.delete_and_close()
-                    raise e
-            except Exception:
-                yield self.delete_and_close()
-                raise
-            else:
                 break
+            except TimeoutError:
+                # we may need to reset the connection and try again
+                yield self._sender.close_connections()
+            except Exception:
+                pass
         try:
             stdout, stderr = yield self.receive()
         except TimeoutError:
             # close_connections done in receive() for TimeoutError
             raise
+        except RequestError:
+            yield self._sender.send_request('delete', shell_id=self._shell_id)
+            yield self._sender.close_connections()
+            self._shell_id = None
+            defer.returnValue(CommandResponse([], [], 0))
         try:
             yield self._sender.send_request(
                 'signal',
@@ -304,7 +403,9 @@ class LongRunningCommand(object):
                 signal_code=c.SHELL_SIGNAL_TERMINATE)
         except RequestError:
             pass
-        yield self.delete_and_close()
+        yield self._sender.send_request('delete', shell_id=self._shell_id)
+        yield self._sender.close_connections()
+        self._shell_id = None
         defer.returnValue(CommandResponse(stdout, stderr, self._exit_code))
 
 
