@@ -40,7 +40,7 @@ _NANOSECONDS_PATTERN = re.compile(r'\.(\d{6})(\d{3})')
 _REQUEST_TEMPLATE_NAMES = (
     'enumerate', 'pull',
     'create', 'command', 'send', 'receive', 'signal', 'delete',
-    'subscribe', 'event_pull', 'unsubscribe')
+    'subscribe', 'event_pull', 'unsubscribe', 'enum_shells', 'pull_shells')
 _REQUEST_TEMPLATE_DIR = os.path.join(
     os.path.dirname(os.path.abspath(__file__)), 'request')
 _REQUEST_TEMPLATES = {}
@@ -624,11 +624,8 @@ class RequestSender(object):
 
         @defer.inlineCallbacks
         def reset_agent_resend(sender, request, body_producer):
-            sender.agent = _get_agent()
-            if sender.gssclient is not None:
-                # do some cleanup first.  memory leaks were occurring
-                sender.gssclient.cleanup()
-                sender.gssclient = None
+            yield self.close_connections()
+            if sender.is_kerberos():
                 try:
                     yield sender._set_url_and_headers()
                     encrypted_request = sender.gssclient.encrypt_body(request)
@@ -666,8 +663,8 @@ class RequestSender(object):
                         if auth_details:
                             yield self.gssclient._step(auth_details)
                     except kerberos.GSSError as e:
-                        msg ="HTTP Unauthorized received.  "\
-                        "Kerberos error code {0}: {1}.".format(e.args[1][1],e.args[1][0])
+                        msg = "HTTP Unauthorized received.  "\
+                            "Kerberos error code {0}: {1}.".format(e.args[1][1], e.args[1][0])
                         raise Exception(msg)
                 raise UnauthorizedError(
                     "HTTP Unauthorized received: Check username and password")
@@ -681,22 +678,27 @@ class RequestSender(object):
                 reader = _ErrorReader()
             response.deliverBody(reader)
             message = yield reader.d
+            if 'maximum number of concurrent operations for this user has been exceeded' in message:
+                message += '  To fix this, increase the MaxConcurrentOperationsPerUser WinRM'\
+                           ' Configuration option to 4294967295 and restart the winrm service.'
             raise RequestError("HTTP status: {0}. {1}".format(
                 response.code, message))
         defer.returnValue(response)
 
+    @defer.inlineCallbacks
     def close_connections(self):
         # close connections
         # return a Deferred()
         if self.agent and hasattr(self.agent, 'closeCachedConnections'):
             # twisted 11 has no return and is part of the Agent
-            return defer.succeed(self.agent.closeCachedConnections())
+            self.agent.closeCachedConnections()
         elif self.agent:
-            # twisted 12 returns a Deferred
-            return self.agent._pool.closeCachedConnections()
-        else:
-            # no agent
-            return defer.succeed(None)
+            # twisted 12 returns a Deferred from the pool
+            yield self.agent._pool.closeCachedConnections()
+        if self.gssclient is not None:
+            self.gssclient.cleanup()
+            self.gssclient = None
+        defer.returnValue(None)
 
 
 class _StringProtocol(Protocol):
@@ -734,7 +736,7 @@ class EtreeRequestSender(object):
                 import xml.dom.minidom
                 xml = xml.dom.minidom.parseString(xml_str)
                 log.debug(xml.toprettyxml())
-            except:
+            except Exception:
                 log.debug('Could not prettify response XML: "{0}"'.format(xml_str))
         defer.returnValue(ET.fromstring(xml_str))
 
